# -*- coding: utf-8 -*-
"""Graphs of consumption data from the Octopus Energy API"""
from statistics import mean
from datetime import time
import numpy
import plotly.graph_objects as go

from octopy_graph.query import Period

DAYS = {1: "Mon", 2: "Tue", 3: "Wed", 4: "Thu", 5: "Fri", 6: "Sat", 7: "Sun"}


def axis(consumptions):
    """Separate a query result into x and y axis

    >>> from collections import namedtuple
    >>> C = namedtuple("C", "interval_start consumption")
    >>> axis([C("a", 1), C("b", 2), C("c", 3)])
    (['c', 'b', 'a'], [3, 2, 1])
    >>> axis([])
    ([], [])

    :param consumptions: a list of Consumption objects
    :return: Tuple (x-axis, y-axis)
    """
    xaxis = [c.interval_start for c in reversed(consumptions)]
    yaxis = [c.consumption for c in reversed(consumptions)]
    return xaxis, yaxis


def axis_weekly_consumption(consumptions):
    """Separate a query result into x and y axis, summed per week

    >>> from collections import namedtuple
    >>> from datetime import date
    >>> C = namedtuple("C", "interval_start consumption")
    >>> axis_weekly_consumption([C(date(2017, 5, 12), 3)])
    (['[17] W19'], [3])
    >>> axis_weekly_consumption([C(date(2017, 5, 12), 3), C(date(2017, 5, 13), 2)])
    (['[17] W19'], [5])
    >>> axis_weekly_consumption([C(date(2017, 5, 12), 3), C(date(2017, 5, 15), 2)])
    (['[17] W19', '[17] W20'], [3, 2])

    :param consumptions: a list of Consumption objects
    :return: Tuple (x-axis, y-axis)
    """
    all_weeks = list({_week_num(c.interval_start) for c in consumptions})
    all_weeks = sorted(all_weeks)
    yaxis = [
        sum(c.consumption for c in consumptions if _week_num(c.interval_start) == week)
        for week in all_weeks
    ]
    return all_weeks, yaxis


def axis_average_day(consumptions):
    """Separate a query result into x and y axis, averaged per day of the week

    >>> from collections import namedtuple
    >>> from datetime import date
    >>> C = namedtuple("C", "interval_start consumption")
    >>> axis_average_day([C(date(2017, 5, 12), 3)])
    ([5], [3])
    >>> axis_average_day([C(date(2017, 5, 12), 3), C(date(2017, 5, 19), 5)])
    ([5], [4])
    >>> axis_average_day([C(date(2017, 5, 12), 3), C(date(2017, 5, 15), 4)])
    ([1, 5], [4, 3])

    :param consumptions: a list of Consumption objects
    :return: Tuple (x-axis, y-axis)
    """
    all_days = list({c.interval_start.isoweekday() for c in consumptions})
    all_days = sorted(all_days)
    yaxis = [
        mean(
            c.consumption for c in consumptions if c.interval_start.isoweekday() == day
        )
        for day in all_days
    ]
    return all_days, yaxis


def axis_average_hour(consumptions):
    """Separate a query result into x and y axis, averaged per hour

    >>> from collections import namedtuple
    >>> from datetime import time
    >>> C = namedtuple("C", "interval_start consumption")
    >>> axis_average_hour([C(time(12, 10, 20), 3)])
    ([12], [3])
    >>> axis_average_hour([C(time(12, 10, 20), 3), C(time(12, 49, 59), 7)])
    ([12], [5])
    >>> axis_average_hour([C(time(12, 00, 00), 3), C(time(11, 59, 59), 4)])
    ([11, 12], [4, 3])

    :param consumptions: a list of Consumption objects
    :return: Tuple (x-axis, y-axis)
    """
    all_hours = list({c.interval_start.hour for c in consumptions})
    all_hours = sorted(all_hours)
    yaxis = [
        mean(c.consumption for c in consumptions if c.interval_start.hour == hour)
        for hour in all_hours
    ]
    return all_hours, yaxis


def average_hourly_consumption(meter_tool):
    """Bar chart showing average hourly consumption

    :param meter_tool: a MeterQueryTool
    """
    elec_hour = meter_tool.electricity_consumption(period=Period.HOUR)
    hours, electricity = axis_average_hour(elec_hour)
    gas_hour = meter_tool.gas_consumption(period=Period.HOUR)
    _, gas = axis_average_hour(gas_hour)
    time_axis = numpy.array([time(hour=h).strftime("%H:%M") for h in hours])

    fig = go.Figure(
        data=[
            go.Bar(name="Electricity", x=time_axis, y=electricity),
            go.Bar(name="Gas", x=time_axis, y=gas),
        ]
    )

    fig.update_layout(
        barmode="group",
        title=go.layout.Title(text="Average Hourly Consumption", xref="paper", x=0),
        legend=dict(x=0.6, y=1.15),
        legend_orientation="h",
        yaxis=go.layout.YAxis(
            title=go.layout.yaxis.Title(
                text="kWh",
                font=dict(family="Courier New, monospace", size=18, color="#7f7f7f"),
            )
        ),
    )

    return fig


def average_daily_consumption(meter_tool):
    """Bar chart showing average consumption per day of the week

    :param meter_tool: a MeterQueryTool
    """
    elec_hour = meter_tool.electricity_consumption(period=Period.DAY)
    days, electricity = axis_average_day(elec_hour)
    gas_hour = meter_tool.gas_consumption(period=Period.DAY)
    _, gas = axis_average_day(gas_hour)
    time_axis = numpy.array([DAYS[d] for d in days])

    fig = go.Figure(
        data=[
            go.Bar(name="Electricity", x=time_axis, y=electricity),
            go.Bar(name="Gas", x=time_axis, y=gas),
        ]
    )

    fig.update_layout(
        barmode="group",
        title=go.layout.Title(text="Average Daily Consumption", xref="paper", x=0),
        legend=dict(x=0.6, y=1.15),
        legend_orientation="h",
        yaxis=go.layout.YAxis(
            title=go.layout.yaxis.Title(
                text="kWh",
                font=dict(family="Courier New, monospace", size=18, color="#7f7f7f"),
            )
        ),
    )

    return fig


def daily_consumption(meter_tool, date):
    """Bar chart showing consumption on a specific day

    :param meter_tool: a MeterQueryTool
    :param date: A datetime representing the date to graph
    """
    date_name = date.strftime("%d/%m/%Y")
    elec_hour = meter_tool.electricity_consumption()
    times, electricity = axis(
        [c for c in elec_hour if c.interval_start.date() == date.date()]
    )
    gas_hour = meter_tool.gas_consumption()
    _, gas = axis([c for c in gas_hour if c.interval_start.date() == date.date()])
    time_axis = numpy.array([t.strftime("%H:%M") for t in times])

    fig = go.Figure(
        data=[
            go.Bar(name="Electricity", x=time_axis, y=electricity),
            go.Bar(name="Gas", x=time_axis, y=gas),
        ]
    )

    fig.update_layout(
        barmode="group",
        title=go.layout.Title(text=f"Consumption on {date_name}", xref="paper", x=0),
        legend=dict(x=0.6, y=1.15),
        legend_orientation="h",
        yaxis=go.layout.YAxis(
            title=go.layout.yaxis.Title(
                text="kWh",
                font=dict(family="Courier New, monospace", size=18, color="#7f7f7f"),
            )
        ),
    )

    return fig


def weekly_consumption(meter_tool, date):
    """Bar chart showing consumption on a week day

    :param meter_tool: a MeterQueryTool
    :param date: A datetime for a date in the desired week
    """
    week_name = f"Week {date.isocalendar()[1]} {date.year}"
    elec_hour = meter_tool.electricity_consumption(period=Period.DAY)
    days, electricity = axis(
        [c for c in elec_hour if _week_num(c.interval_start) == _week_num(date)]
    )
    gas_hour = meter_tool.gas_consumption(period=Period.DAY)
    _, gas = axis(
        [c for c in gas_hour if _week_num(c.interval_start) == _week_num(date)]
    )
    time_axis = numpy.array([DAYS[d.isoweekday()] for d in days])

    fig = go.Figure(
        data=[
            go.Bar(name="Electricity", x=time_axis, y=electricity),
            go.Bar(name="Gas", x=time_axis, y=gas),
        ]
    )

    fig.update_layout(
        barmode="group",
        title=go.layout.Title(text=f"Consumption on {week_name}", xref="paper", x=0),
        legend=dict(x=0.6, y=1.15),
        legend_orientation="h",
        yaxis=go.layout.YAxis(
            title=go.layout.yaxis.Title(
                text="kWh",
                font=dict(family="Courier New, monospace", size=18, color="#7f7f7f"),
            )
        ),
    )

    return fig


def average_weekly_consumption(meter_tool):
    """Bar chart showing weekly consumption

    :param meter_tool: a MeterQueryTool
    """
    elec_hour = meter_tool.electricity_consumption(period=Period.WEEK)
    weeks, electricity = axis_weekly_consumption(elec_hour)
    gas_hour = meter_tool.gas_consumption(period=Period.WEEK)
    _, gas = axis_weekly_consumption(gas_hour)
    time_axis = numpy.array(weeks)

    fig = go.Figure(
        data=[
            go.Bar(name="Electricity", x=time_axis, y=electricity),
            go.Bar(name="Gas", x=time_axis, y=gas),
        ]
    )

    fig.update_layout(
        barmode="group",
        title=go.layout.Title(text="Weekly Consumption", xref="paper", x=0),
        legend=dict(x=0.6, y=1.15),
        legend_orientation="h",
        yaxis=go.layout.YAxis(
            title=go.layout.yaxis.Title(
                text="kWh",
                font=dict(family="Courier New, monospace", size=18, color="#7f7f7f"),
            )
        ),
    )

    return fig


def _week_num(date):
    """Return the week number for a specific date

    >>> from datetime import date
    >>> _week_num(date(2018, 10, 10,))
    '[18] W41'
    >>> _week_num(date(2019, 1, 31,))
    '[19] W5'

    :param date: a datetime object
    """
    week_num = date.isocalendar()[1]
    year = date.year - 2000
    return f"[{year}] W{week_num}"



# -*- coding: utf-8 -*-
"""Generate the graphs supported by the tool"""
from datetime import datetime, timedelta
import argparse
import sys

import octopy_graph as og


def get_parser():
    """The command line parser"""
    parser = argparse.ArgumentParser(description="GitDraw")
    parser.add_argument("-k", "--apikey", help="The API Key", required=True)
    parser.add_argument(
        "-g", "--gasserial", help="The gas meters serial number", required=True
    )
    parser.add_argument("-r", "--mprn", help="The gas meters MPRN", required=True)
    parser.add_argument(
        "-e",
        "--electricserial",
        help="The electricity meters serial number",
        required=True,
    )
    parser.add_argument(
        "-a", "--mpan", help="The electricity meters MPAN", required=True
    )
    return parser


def get_tool(key, gas_serial, elec_serial, mprn, mpan):
    """Get a MeterQueryTool"""
    g_meter = og.GasMeter(mprn, gas_serial)
    e_meter = og.ElectricityMeter(mpan, elec_serial)
    return og.MeterQueryTool(key, g_meter, e_meter)


def main():
    """Main program path"""
    parser = get_parser()
    args = parser.parse_args()
    tool = get_tool(
        args.apikey, args.gasserial, args.electricserial, args.mprn, args.mpan
    )

    fig1 = og.average_hourly_consumption(tool)
    fig2 = og.average_daily_consumption(tool)
    fig3 = og.average_weekly_consumption(tool)
    fig4 = og.daily_consumption(tool, datetime.now() - timedelta(days=1))
    fig5 = og.weekly_consumption(tool, datetime.now() - timedelta(days=7))

    fig1.show()
    fig3.show()
    fig2.show()
    fig4.show()
    fig5.show()


if __name__ == "__main__":
    sys.exit(main())



# -*- coding: utf-8 -*-
"""Graphing data from Octopus Energy API"""
from octopy_graph.query import MeterQueryTool, GasMeter, ElectricityMeter, Period
from octopy_graph.graph import (
    average_hourly_consumption,
    average_daily_consumption,
    average_weekly_consumption,
    daily_consumption,
    weekly_consumption,
)



import os
import setuptools

# Crete new tag & update download url & version
# Run:
# venv\Scripts\python.exe setup.py sdist bdist_wheel
# venv\Scripts\twine.exe upload dist/*

if not os.environ.get("CI_COMMIT_TAG"):
    exit()

version = os.environ["CI_COMMIT_TAG"]

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="octopy-graph",
    packages=["octopy_graph"],
    version=version,
    license="MIT",
    description="A tool for graphing data from the Octopus Energy API",
    long_description=long_description,
    long_description_content_type="text/markdown",
    entry_points={"console_scripts": ["octopy-graph=octopy_graph.__main__:main"]},
    author="Samuel Broster",
    author_email="s.h.broster@gmail.com",
    url="https://gitlab.com/broster/octopy-graph",
    keywords=["Octopus Energy", "plotly", "graph"],
    install_requires=["plotly", "octopy-api", "numpy"],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.7",
    ],
)



# -*- coding: utf-8 -*-
"""Wrappers around the Octopus Energy API swagger client"""
import re
from enum import Enum
from functools import partial
from datetime import datetime
from dataclasses import dataclass
from swagger_client import ApiClient, DefaultApi, Configuration


NUM_RESULTS = 25000
DATE_FMT = "%Y-%m-%dT%H:%M:%S%z"
R_DATE_FMT = re.compile(
    "([0-9]{4}-[0-9]{2}-[0-9]{2}T)"
    "([0-9]{2}:[0-9]{2}:[0-9]{2})"
    "(\+|\-)([0-9]{2}):([0-9]{2})"  # pylint: disable=W1401
)


@dataclass
class GasMeter:
    """Information uniquely identifying a gas meter"""

    mprn: str
    serial_number: str


@dataclass
class ElectricityMeter:
    """Information uniquely identifying an electricity meter"""

    mpan: str
    serial_number: str


class Period(Enum):
    """A time period used by the API"""

    HOUR = "hour"
    DAY = "day"
    WEEK = "week"
    MONTH = "month"
    QUARTER = "quarter"


class MeterQueryTool:
    """Query a Gas and Electricity meter using the API

    :param api_key: The API key
    :param gas_meter: A GasMeter
    :param electricity_meter: An ElectricityMeter
    """

    def __init__(self, api_key, gas_meter, electricity_meter):
        self.api = setup_api(api_key)
        self.gas_meter = gas_meter
        self.electricity_meter = electricity_meter

    def gas_consumption(self, period=None):
        """Get details on gas consumption

        >>> from minimock import Mock
        >>> DefaultApi.gas_meter_points_mprn_meters_serial_number_consumption_get = \
          Mock('DefaultApi.gas_meter_points_mprn_meters_serial_number_consumption_get')
        >>> Result = Mock("Result")
        >>> Result.results = []
        >>> DefaultApi.gas_meter_points_mprn_meters_serial_number_consumption_get \
          .mock_returns = Result
        >>> e = ElectricityMeter("1", "2")
        >>> g = GasMeter("3", "4")
        >>> tool = MeterQueryTool(1, g, e)
        >>> tool.gas_consumption()
        Called DefaultApi.gas_meter_points_mprn_meters_serial_number_consumption_get(
            '3',
            '4',
            page_size=25000)
        []

        :param period: a Period enum, otherwise defaults to half hourly info
        :return: a swagger_client.Consumption
        """
        query = partial(
            self.api.gas_meter_points_mprn_meters_serial_number_consumption_get,
            self.gas_meter.mprn,
            self.gas_meter.serial_number,
            page_size=NUM_RESULTS,
        )
        if period:
            query = partial(query, group_by=period.value)
        return _consumption_query(query)

    def electricity_consumption(self, period=None):
        """Get details on electricity consumption

        >>> from minimock import Mock
        >>> DefaultApi.electricity_meter_points_mpan_meters_serial_number_consumption_get \
          = Mock('DefaultApi.electricity_meter_points_mpan_meters_serial_number_consumption_get')
        >>> Result = Mock("Result")
        >>> Result.results = []
        >>> DefaultApi.electricity_meter_points_mpan_meters_serial_number_consumption_get \
          .mock_returns = Result
        >>> e = ElectricityMeter("1", "2")
        >>> g = GasMeter("3", "4")
        >>> tool = MeterQueryTool(1, g, e)
        >>> tool.electricity_consumption()
        Called DefaultApi.electricity_meter_points_mpan_meters_serial_number_consumption_get(
            '1',
            '2',
            page_size=25000)
        []

        :param period: a Period enum, otherwise defaults to half hourly info
        :return: an instance of swagger_client.Consumption
        """
        query = partial(
            self.api.electricity_meter_points_mpan_meters_serial_number_consumption_get,
            self.electricity_meter.mpan,
            self.electricity_meter.serial_number,
            page_size=NUM_RESULTS,
        )
        if period:
            query = partial(query, group_by=period.value)
        return _consumption_query(query)


def _consumption_query(query):
    """Handle issues a 'consumption' API query

    :param query: A function representing the query to issue
    """
    result = query().results
    for consumption in result:
        consumption.interval_end = _fix_date(consumption.interval_end)
        consumption.interval_start = _fix_date(consumption.interval_start)
    return result


def _fix_date(date):
    """Turn a date string into a datetime object

    >>> _fix_date("2018-05-19T23:10:01Z")
    datetime.datetime(2018, 5, 19, 23, 10, 1, tzinfo=datetime.timezone.utc)
    >>> _fix_date("2018-05-19T23:10:01+01:00")
    datetime.datetime(2018, 5, 19, 23, 10, 1, tzinfo=\
datetime.timezone(datetime.timedelta(0, 3600)))

    :param date: The date to fix
    """
    if date[-1] == "Z":
        date = date[:-1] + "+00:00"
    fixed_format = R_DATE_FMT.sub(r"\1\2\3\4\5", date)
    return datetime.strptime(fixed_format, DATE_FMT)


def setup_api(api_key):
    """Setup the Octopus Energy API

    >>> setup_api("mykey")
    <swagger_client.api.default_api.DefaultApi object at ...

    :param api_key: The API key"""
    conf = Configuration()
    conf.username = api_key
    return DefaultApi(ApiClient(conf))
